"""Post analysis for Blobby3D output.

To be used after running Blobby3D to organise the Blobby3D output.


Attributes
----------
Classes: Blobby3D

@author: mathewvaridel

"""

from pathlib import Path
import numpy as np
import pandas as pd

from .meta import Metadata


class PostBlobby3D:

    def __init__(
            self, samples_path, data_path, var_path, metadata_path,
            save_maps=True, save_precon=True, save_con=True,
            nlines=1, nsigmad=2):
        """Blobby3D postprocess object.

        This provides a read and storage object for analysis.

        Parameters
        ----------
        samples_path : str or pathlib object
            Sample path. This can be either DNest4 samples or posterior
            samples.
        data_path : str or pathlib object
            Data cube path. Data cubes are assumed to be in whitespace
            separated text files. The values should be in row-major order.
        var_path : str or pathlib object
            Variance cube path. Variance cubes are assumed to be in whitespace
            separated text files. The values should be in row-major order.
        metadata_path : str or pathlib object
            Metadata file path. This path records the coordinates for your
            cube.
        save_maps : bool, optional
            Maps were saved to sampled by Blobby3D. The default is True.
        save_precon : bool, optional
            Preconvolved cubes were saved to samples by Blobby3D. The default
            is True.
        save_con : bool, optional
            Convolved cubes were saved to samples by Blobby3D. The default is
            True.
        nlines : int, optional
            Number of lines modelled during run. A coupled line count as one.
            Therefore modelling H-alpha and [NII] at 6548.1 A and 6583.1 A
            with a fixed ratio is counted as two lines in total. The default is
            1.
        nsigmad : int, optional
            The degree to which white and shot noise are modelled. The default
            is 2.

        Attributes
        ----------
        nsamples : int
            Number of samples recorded.
        max_blobs : int
            Maximum number of blobs allowed in run.
        metadata : Metadata object
            See Metadata class docstring for details.
        data : np.ndarray
        var : np.ndarray
        maps : np.ndarray
            2D maps of flux per line, velocity, and velocity dispersion for
            each sample..
        precon_cubes : np.ndarray
            Preconvolved model cubes for each sample.
        con_cubes : np.ndarray
            Convolved model cubes for each sample.
        global_param : pd.DataFrame
            Global parameters.
        blob_param :
            Parameters for each blob.

        """
        self._posterior_path = Path(samples_path)
        self._data_path = Path(data_path)
        self._var_path = Path(var_path)
        self._metadata_path = Path(metadata_path)
        self._nlines = nlines
        self._nsigmad = nsigmad

        # import data
        self.metadata = Metadata(self._metadata_path)
        self.data = np.loadtxt(data_path).reshape(self.metadata.naxis)
        self.var = np.loadtxt(var_path).reshape(self.metadata.naxis)

        # posterior samples
        samples = np.atleast_2d(np.loadtxt(samples_path))
        self.nsamples = samples.shape[0]

        if save_maps:
            self.maps = np.zeros((
                    self.nsamples,
                    self._nlines+2,
                    *self.metadata.naxis[:2]))
            map_shp = self.metadata.naxis[:2].prod()
            for s in range(self.nsamples):
                # Flux
                for ln in range(self._nlines):
                    self.maps[s, ln, :, :] = samples[
                            s, ln*map_shp:(1+ln)*map_shp
                            ].reshape(self.metadata.naxis[:2])

                # LoS velocity
                self.maps[s, self._nlines, :, :] = samples[
                        s, self._nlines*map_shp:(1+self._nlines)*map_shp
                        ].reshape(self.metadata.naxis[:2])

                # LoS vdisp
                self.maps[s, 1+self._nlines, :, :] = samples[
                        s, (1+self._nlines)*map_shp:(2+self._nlines)*map_shp
                        ].reshape(self.metadata.naxis[:2])

        if save_con:
            self.precon_cubes = np.zeros((self.nsamples, *self.metadata.naxis))
            st = save_maps*(2+self._nlines)*map_shp
            for s in range(self.nsamples):
                self.precon_cubes[s, :, :, :] = samples[
                        s, st:st+self.metadata.sz].reshape(self.metadata.naxis)

        if save_precon:
            self.con_cubes = np.zeros((self.nsamples, *self.metadata.naxis))
            st = save_maps*(2+self._nlines)*map_shp
            st += save_precon*self.metadata.sz
            for s in range(self.nsamples):
                self.con_cubes[s, :, :, :] = samples[
                        s, st:st+self.metadata.sz].reshape(self.metadata.naxis)

        # blob parameters
        st = save_maps*(2+self._nlines)*map_shp
        st += self.metadata.sz*(save_precon + save_precon)
        self.max_blobs = int(samples[0, st+1])

        global_names = ['WD', 'RADMIN', 'RADMAX', 'QMIN']
        for i in range(self._nlines):
            global_names += ['FLUX%iMU' % (i), 'FLUX%iSD' % (i)]
        global_names += [
                'NUMBLOBS',
                'XC', 'YC',
                'DISKFLUX', 'DISKMU',
                'VSYS', 'VMAX', 'VSLOPE', 'VGAMMA', 'VBETA'
                ]
        global_names += ['VDISP%i' % (i) for i in range(self._nsigmad)]
        global_names += ['INC', 'PA', 'SIGMA0', 'SIGMA1']
        global_param = np.concatenate((
                samples[:, st+2:st+7+2*self._nlines],
                samples[:, -13-self._nsigmad:]), axis=1)
        self.global_param = pd.DataFrame(global_param, columns=global_names)
        self.global_param.index.name = 'SAMPLE'

        if self.max_blobs > 0:
            blob_names = ['RC', 'THETAC', 'W', 'Q', 'PHI']
            blob_names += ['FLUX%i' % (i) for i in range(self._nlines)]
            n_bparam = len(blob_names)
            self.blob_param = np.zeros(
                    (self.nsamples*self.max_blobs, n_bparam))
            st_bparam = st+7+2*self._nlines
            end_bparam = -13-self._nsigmad
            for s in range(self.nsamples):
                row_st = self.max_blobs*s
                row_end = self.max_blobs*(s + 1)
                sblob_param = samples[s, st_bparam:end_bparam]
                sblob_param = sblob_param.reshape(
                        (self.max_blobs, n_bparam), order='F')
                self.blob_param[row_st:row_end, :] = sblob_param

            self.blob_param = pd.DataFrame(self.blob_param, columns=blob_names)

            self.blob_param['SAMPLE'] = np.repeat(
                    np.arange(self.nsamples), self.max_blobs)
            self.blob_param['BLOB'] = np.tile(
                    np.arange(self.max_blobs), self.nsamples)
            self.blob_param.set_index(['SAMPLE', 'BLOB'], inplace=True)
            self.blob_param = self.blob_param[self.blob_param['RC'] > 0.0]

    # def plot_global_marginalised(self, save_file=None):
    #     if isinstance(save_file, str):
    #         pdf_file = mpl.backends.backend_pdf.PdfPages(save_file)

    #     for key in self.global_param.keys():
    #         fig = plt.figure()
    #         ax = plt.gca()
    #         ax.hist(self.global_param[key])
    #         ax.set_title(r'%s' % (str(key)))
    #         if save_file:
    #             pdf_file.savefig(fig)
    #             plt.close()

    #     if isinstance(save_file, str):
    #         pdf_file.close()

    # def plot_map(self, ax, map_2d, **kwargs):
    #     """Plot individual map to a given axes object."""
    #     b3dplot.plot_map(ax, map_2d, **kwargs)

    # def setup_comparison_maps(
    #         self, figsize=(10.0, 10.0), log_flux=True, **kwargs):
    #     """Setup comparsion maps for a given sample."""
    #     fig, ax = b3dcomp.setup_comparison_maps(
    #             comp_shape=(2 + self._nlines, 3),
    #             map_shape=self.metadata.naxis[:2],
    #             figsize=figsize,
    #             **kwargs)

    #     return fig, ax

    # def add_comparison_maps(self, maps, ax, col, log_flux=False):
    #     """Add a column worth of maps to the comparison maps."""
    #     for line in range(self._nlines):
    #         if log_flux:
    #             with np.errstate(invalid='ignore', divide='ignore'):
    #                 flux_map = np.log10(maps[line, :, :])
    #                 flux_map[~np.isfinite(flux_map)] = np.nan
    #         else:
    #             flux_map = maps[line, :, :]
    #         self.plot_map(ax[line][col], flux_map, cmap=b3dplot.cmap.flux)
    #     self.plot_map(
    #             ax[self._nlines][col], maps[self._nlines, :, :],
    #             cmap=b3dplot.cmap.v)
    #     self.plot_map(
    #             ax[self._nlines+1][col], maps[self._nlines+1, :, :],
    #             cmap=b3dplot.cmap.vdisp)

    # def add_comparison_residuals(self, maps, ax, col, log_flux=False):
    #     """Add a column worth of maps to the comparison maps."""
    #     for line in range(self._nlines):
    #         if log_flux:
    #             flux_map = np.log10(maps[line, :, :])
    #         else:
    #             flux_map = maps[line, :, :]
    #         self.plot_map(ax[line][col], flux_map, cmap=b3dplot.cmap.residuals)
    #     self.plot_map(
    #             ax[self._nlines][col], maps[self._nlines, :, :],
    #             cmap=b3dplot.cmap.residuals)
    #     self.plot_map(
    #             ax[self._nlines+1][col], maps[self._nlines+1, :, :],
    #             cmap=b3dplot.cmap.residuals)

    # def update_comparison_clim(
    #         self, ax, cax,
    #         pct=100.0, vlim=None, absolute=False, **cb_kwargs):
    #     """Update each row colour scale for comparison plots."""
    #     if isinstance(ax, (list, tuple)):
    #         data = np.array([a.images[0]._A.data for a in ax])
    #         clim = b3dcomp.map_limits(data, pct, vlim, absolute)
    #         for a in ax:
    #             a.images[0].set_clim(clim)
    #         b3dcomp.colorbar(ax[-1].images[0], cax, clim, **cb_kwargs)
    #     else:
    #         data = ax.images[0]._A.data
    #         clim = b3dcomp.map_limits(data, pct, vlim, absolute)
    #         ax.images[0].set_clim(clim)
    #         b3dcomp.colorbar(ax.images[0], cax, clim, **cb_kwargs)

    # def update_comparison_colorbar(self, ax, **cb_kwargs):
    #     """
    #     Update colorbars in accordance with corresponding maps.

    #     ax : list of matplotlib.axes
    #         The right-most axis is the colorbar. The preceding maps are the
    #         maps with the corresponding clim.
    #     """
    #     clims = [a.images[0].get_clim() for a in ax[:-1]]
    #     for clim in clims:
    #         assert clim == clims[0]

    #     b3dcomp.colorbar(ax[-1], clims[0], **cb_kwargs)

    # def update_comparison_mask(self, mask, ax):
    #     """
    #     Update mask for maps.

    #     Parameters
    #     ----------
    #     ax : matplotlib.axes._subplots.AxesSubplot
    #     """
    #     if isinstance(ax, (list, tuple)):
    #         for i, a in enumerate(ax):
    #             a.images[0]._A = np.ma.masked_where(mask, a.images[0]._A)
    #     else:
    #         ax.images[0]._A = np.ma.masked_where(mask, ax.images[0]._A)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Metadata.

Organisation of the Blobby3D metadata that describes the coordinates.

@author: Mathew Varidel
"""

from pathlib import Path
import numpy as np


class Metadata:

    def __init__(self, metadata_path):
        """Metadata oranisation object.

        Parameters
        ----------
        metadata_path : str or pathlib.Path
            DESCRIPTION.

        Returns
        -------
        None.

        """
        metadata = np.loadtxt(Path(metadata_path))
        self.naxis = metadata[:3].astype(int)
        self.sz = self.naxis.prod()
        self.x_lim = metadata[3:5]
        self.y_lim = metadata[5:7]
        self.r_lim = metadata[7:9]
        self.dx = float(np.diff(self.x_lim)/self.naxis[1])
        self.dy = float(np.diff(self.y_lim)/self.naxis[0])
        self.dr = float(np.diff(self.r_lim)/self.naxis[2])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Plotting routines.

@author: Mathew Varidel

"""

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


class cmap:
    flux = 'Greys_r'
    v = 'RdYlBu_r'
    vdisp = 'YlOrBr'
    residuals = 'RdYlBu_r'


def plot_map(
        ax, map_2d,
        title=None,
        xlabel=None,
        ylabel=None,
        xticks=False,
        yticks=False,
        xlim=None,
        ylim=None,
        colorbar=False,
        colorbar_cax=None,
        cbar_label=None,
        clim=None,
        cbarticks=None,
        cbar_nticks=5,
        logscale=False,
        cmap=None,
        title_fontsize='large',
        label_fontsize='large',
        tick_fontsize='large',
        cb_label_fontsize='large',
        cb_tick_fontsize='large',
        fwhm=None,
        aspect=None,
        mask=None):
    """Plot singular 2d map to an axes."""
    naxis = map_2d.shape

    if clim is not None:
        if clim[0] is None:
            clim[0] = np.nanmin(map_2d)
        if clim[1] is None:
            clim[1] = np.nanmax(map_2d)
    else:
        clim = [np.nanmin(map_2d), np.nanmax(map_2d)]

    if logscale:
        norm = mpl.colors.LogNorm(vmin=clim[0], vmax=clim[1])
    else:
        norm = None

    ax.imshow(
        map_2d,
        origin='lower',
        interpolation='nearest',
        norm=norm,
        vmin=clim[0], vmax=clim[1],
        cmap=cmap,
        aspect=aspect
        )

    ax.tick_params(axis='both', direction='in', pad=4.0)

    if isinstance(title, str):
        ax.set_title(title, fontsize=title_fontsize)

    if isinstance(xlabel, str):
        ax.set_xlabel(xlabel, fontsize=label_fontsize)

    if isinstance(ylabel, str):
        ax.set_ylabel(ylabel, fontsize=label_fontsize)

    set_ticks(
            ax.xaxis,
            ticks=xticks,
            naxis=naxis[1],
            lim=xlim,
            tick_fontsize=tick_fontsize)

    set_ticks(
            ax.yaxis,
            ticks=yticks,
            naxis=naxis[0],
            lim=ylim,
            tick_fontsize=tick_fontsize)



    # XTicks
#    # TODO: Combine XTicks/YTicks settings
#    if not xticks:
#        ax.set_xticks([])
#    elif xticks == 'all':
#        x_dist = np.diff(self.x_lim)[0]/2.0
#        ax.set_xticks([-0.5, (naxis[1]-1.0)/2.0, naxis[1]-0.5])
#        xtick_values = [-round(x_dist, 1), 0.0, round(x_dist, 1)]
#        ax.set_xticklabels(xtick_values, fontsize=tick_fontsize)
#    elif xticks == 'upper':
#        x_dist = np.diff(self.x_lim)[0]/2.0
#        ax.set_xticks([(naxis[1]-1.0)/2.0, naxis[1]-0.5])
#        xtick_values = [0.0, round(x_dist, 1)]
#        ax.set_xticklabels(xtick_values, fontsize=tick_fontsize)
#    elif xticks == 'lower':
#        x_dist = np.diff(self.x_lim)[0]/2.0
#        ax.set_xticks([-0.5, (naxis[1]-1.0)/2.0])
#        xtick_values = [-round(x_dist, 1), 0.0]
#        ax.set_xticklabels(xtick_values, fontsize=tick_fontsize)

#    # YTicks
#    if not yticks:
#        ax.set_yticks([])
#    elif yticks == 'all':
#        y_dist = np.diff(self.y_lim)[0]/2.0
#        ax.set_yticks([-0.5, (naxis[0]-1.0)/2.0, naxis[0]-0.5])
#        ytick_values = [-round(y_dist, 1), 0.0, round(y_dist, 1)]
#        ax.set_yticklabels(ytick_values, fontsize=tick_fontsize)
#    elif yticks == 'upper':
#        y_dist = np.diff(self.y_lim)[0]/2.0
#        ax.set_yticks([(naxis[0]-1.0)/2.0, naxis[0]-0.5])
#        ytick_values = [0.0, round(y_dist, 1)]
#        ax.set_yticklabels(ytick_values, fontsize=tick_fontsize)
#    elif yticks == 'lower':
#        y_dist = np.diff(self.y_lim)[0]/2.0
#        ax.set_yticks([-0.5, (naxis[0]-1.0)/2.0])
#        ytick_values = [-round(y_dist, 1), 0.0]
#        ax.set_yticklabels(ytick_values, fontsize=tick_fontsize)

    if colorbar:
        if colorbar_cax is None:
            divider = make_axes_locatable(ax)
            colorbar_cax = divider.append_axes(
                    'right', size='10%', pad=0.03)

        plot_colorbar(
                cax=colorbar_cax,
                clim=clim,
                label=cbar_label,
                label_fontsize=cb_label_fontsize,
                cbarticks=cbarticks,
                cbar_nticks=cbar_nticks,
                tick_fontsize=cb_tick_fontsize,
                cmap=cmap
                )

    if fwhm is not None:
        circle = plt.Circle(
                (1.1*fwhm, 1.1*fwhm),
                radius=fwhm,
                fill=False, edgecolor='r', linewidth=1.0)
        ax.add_artist(circle)


def set_ticks(ax, naxis, ticks=False, lim=None, tick_fontsize='large'):
    if not ticks:
        ax.set_ticks([])
    elif ticks == 'all':
        dist = np.diff(lim)[0]/2.0
        ax.set_ticks([-0.5, (naxis-1.0)/2.0, naxis-0.5])
        tick_values = [-round(dist, 1), 0.0, round(dist, 1)]
        ax.set_ticklabels(tick_values, fontsize=tick_fontsize)
    elif ticks == 'upper':
        dist = np.diff(lim)[0]/2.0
        ax.set_ticks([(naxis-1.0)/2.0, naxis-0.5])
        tick_values = [0.0, round(dist, 1)]
        ax.set_ticklabels(tick_values, fontsize=tick_fontsize)
    elif ticks == 'lower':
        dist = np.diff(lim)[0]/2.0
        ax.set_ticks([-0.5, (naxis-1.0)/2.0])
        tick_values = [-round(dist, 1), 0.0]
        ax.set_ticklabels(tick_values, fontsize=tick_fontsize)
    else:
        raise ValueError('xticks must be False, all, upper, lower.')


def plot_colorbar(
        cax,
        mappable=None,
        label=None,
        xlabel=None,
        ylabel=None,
        clim=None,
        cbarticks=None,
        cbar_nticks=5,
        logscale=False,
        cmap=None,
        label_fontsize='large',
        tick_fontsize='large'):
    """
    Plot colorbar.

    Parameters
    ----------
    cax : matplotlib.pyplot.axes
        Colorbar axis.
    mappable : matplotlib.cm.ScalarMappable, default None
        Mappable object -- usually image. Default None uses a linear
        mappable between limits constructed by clim.
    title : str, default None
    xlabel : str, default None
    ylabel : str, default None
    clim : list, default None
        Color limits. Default uses mappable to construct color limits.
    cbarticks : list, default None
        cbarticks[0] corresponds to ticks.
        cbarticks[1] corresponds to tick labels.
    cbar_nticks : int, default 5
        Number of colorbar ticks if cbarticks is None.
    logscale : bool, default False
        Not allowed at this time.
    cmap : Matplotlib.colormap instance
    label_fontsize : matplotlib fontsize property
    tick_fontisze : matplotlib fontsize property
    """
    assert (
            (mappable is not None)
            or (mappable is None and clim is not None)
            )

    # Deal with limits
    if clim is not None:
        if clim[0] is None:
            clim[0] = np.nanmin(mappable)
        if clim[1] is None:
            clim[1] = np.nanmax(mappable)
    else:
        clim = [
                np.nanmin(mappable),
                np.nanmax(mappable)
                ]

    # logscale
    if logscale:
        cb_norm = mpl.colors.LogNorm(vmin=clim[0], vmax=clim[1])

        # tick formatter
        def logformatter(x, pos):
            value = np.exp(np.log(clim[0]) + x*np.log(clim[1]/clim[0]))
            value = round(value, -int(np.floor(np.log10(abs(value)))))
            return '%i' % (value)
    else:
        cb_norm = mpl.colors.Normalize(vmin=clim[0], vmax=clim[1])

    # Construct default mappable
    if mappable is None:
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=cb_norm)
        sm.set_array([])

    # Construct colorbar
    cb = plt.colorbar(sm, cax=cax)

    if label is not None:
        cb.set_label(label=label, fontsize=label_fontsize)

    cb.set_clim(clim[0], clim[1])

    # Set ticks
    if isinstance(cbarticks, list):
        assert len(cbarticks) == 2
        if logscale:
            def logformatter(x, pos):
                value = np.exp(np.log(clim[0]) + x*np.log(clim[1]/clim[0]))
                value = round(value, -int(np.floor(np.log10(abs(value)))))
                return '%i' % (value)

#                values = [20.0, 50.0, 100.0, 200.0]
            pct = list(map(
                lambda x: np.log(x/clim[0])/np.log(clim[1]/clim[0]),
                cbarticks[0]
                ))

            cb.ax.yaxis.set_ticks(pct)
#                cb.ax.yaxis.set_major_formatter(
#                        mpl.ticker.FuncFormatter(logformatter))
            cb.ax.yaxis.set_major_formatter(
                    mpl.ticker.FixedFormatter(cbarticks[1])
                    )
            cb.ax.yaxis.set_tick_params(
                    labelsize=tick_fontsize,
                    direction='in',
                    pad=4.0
                    )
        else:
            cb.set_ticks(cbarticks[0])
            cb.set_ticklabels(cbarticks[1])
    else:
        if logscale:
            tick_locator = mpl.ticker.LinearLocator(numticks=cbar_nticks)
            cb.locator = tick_locator
        else:
            tick_locator = mpl.ticker.MaxNLocator(nbins=cbar_nticks)
            cb.locator = tick_locator

        cb.ax.tick_params(
            labelsize=tick_fontsize,
            direction='in',
            pad=4.0
            )
        cb.update_ticks()



# -*- coding: utf-8 -*-
"""Physical Constants.

A set of physical constants used across PyBlobby3D.

@author: Mathew Varidel

"""


class PhysicalConstants:
    """Physical constants required for Blobby3D."""
    C = 2.99792458e5  # speed of light (km/s)

    # Vacuum emission line wavelengths.
    HA = 6562.81  # HAlpha (Angstrom)
    N2Low = 6548.1  # Lower NII line
    N2Upp = 6583.1  # Upper NII line



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Construct comparison plots.

@author: Mathew Varidel
"""

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


def setup_comparison_maps(
        comp_shape, map_shape, figsize,
        right_pad=0.1, left_pad=0.05, bottom_pad=0.05,
        xlabel=r'$\Delta$RA ($^{\prime\prime}$)',
        ylabel=r'$\Delta$Dec ($^{\prime\prime}$)'):
    """
    Setup comparison plot.

    This will setup a number of maps of shape 'comp_shape' plus a set of
    residual maps. This is a little fiddly, you may need to play around with
    the pad and figsize attributes to get everything to sit where you want.

    Parameters
    ----------
    comp_shape : list of int
        Comparison map shape (rows, columns). Each row represents a quantity
        of interest, each column represents a map of that quantity for
        comparison.
    map_shape : list of int
        Map array shape in (rows, columns).
    figsize : width, height in inches
        This needs to be defined in order to plot maps with given axis ratio.
    right_pad : float
    left_pad : float
    bottom_pad : float
    xlabel : str
    ylabel : str

    Returns
    -------
    fig : matplotlib.figure.Figure
    ax : nested list of matplotlib.axes._subplots.AxesSubplot
    """
    fig = plt.figure(figsize=figsize)

    maps_tot = comp_shape[1] + 1

    gs_im1 = mpl.gridspec.GridSpec(*comp_shape)
    gs_cb1 = mpl.gridspec.GridSpec(comp_shape[0], 1)
    gs_res = mpl.gridspec.GridSpec(comp_shape[0], 1)
    gs_cb2 = mpl.gridspec.GridSpec(comp_shape[0], 1)

    left_pad = left_pad
    right_pad = right_pad
    bottom_pad = bottom_pad
    cb_width = 0.03
    cb1_right = 0.1
    hspace = 0.0
    wspace = 0.0
    rem_width = (
            1.0 - left_pad - right_pad
            - 2*cb_width - cb1_right - maps_tot*wspace
            )

    im_width = np.round(rem_width/maps_tot, 3)

    # below sets the top such that the aspect ratio for images are correct
    fig_width = fig.get_size_inches()[0]
    im_abswidth = fig_width*im_width

    aim_im_absheight = im_abswidth*map_shape[0]/map_shape[1]

    fig_height = fig.get_size_inches()[1]
    im_height = np.round(aim_im_absheight/fig_height, 3)
    top = comp_shape[0]*im_height + bottom_pad

    # model+data images
    left = left_pad
    right = left + im_width*comp_shape[1] + wspace*comp_shape[1]
    gs_im1.update(
            left=left,
            right=right,
            wspace=wspace,
            bottom=bottom_pad,
            top=top,
            hspace=hspace
            )

    # model+data colorbar
    left = right
    right = left + cb_width
    gs_cb1.update(
            left=left,
            right=right,
            bottom=bottom_pad,
            top=top,
            hspace=hspace
            )

    # residual image
    left = right + cb1_right
    right = left + im_width + wspace
    gs_res.update(
            left=left,
            right=right,
            bottom=bottom_pad,
            top=top,
            hspace=hspace
            )

    # residual colorbar
    left = right
    right = left + cb_width
    gs_cb2.update(
            left=left,
            right=right,
            bottom=bottom_pad,
            top=top,
            hspace=hspace
            )

    ax = []
    for i in range(comp_shape[0]):
        ax_tmp = []

        # Add map axes
        for m in range(comp_shape[1]):
            ax_tmp.append(plt.subplot(gs_im1[i, m]))
        ax_tmp.append(plt.subplot(gs_cb1[i, 0]))

        # Add residual axes
        ax_tmp.append(plt.subplot(gs_res[i, 0]))
        ax_tmp.append(plt.subplot(gs_cb2[i, 0]))

        ax.append(ax_tmp)

    # Plot spatial titles
    fig.text(
        left_pad + 0.5*comp_shape[1]*im_width,
        0.0*bottom_pad,
        xlabel,
        ha='center', fontsize='large'
        )

    fig.text(
        left_pad + (comp_shape[1] + 0.5)*im_width + cb_width + cb1_right,
        0.0*bottom_pad,
        xlabel,
        ha='center', fontsize='large'
        )

    fig.text(
        0.0,
        bottom_pad + 0.5*comp_shape[0]*im_height,
        ylabel,
        va='center', rotation='vertical', fontsize='large'
        )

    return fig, ax


def map_limits(data, pct=100.0, vlim=None, absolute=False):
    """
    Get limits for maps.

    Parameters
    ----------
    data : 2D numpy.array or list of 2D numpy.array
        Map data or list of map data
    pct : float, default 95%
        Percentile to use to calculate limit.
    vlim : list of float, default None
        Absolute limits applied to map.
    absolute : bool, default False
        Whether to calculate limits about 0.
    """
    if isinstance(data, list):
        data_fin = []
        for d in data:
            data_fin.append(d[np.isfinite(d)])
        data_fin = np.array(data_fin)
    else:
        data_fin = data[np.isfinite(data)]

    if vlim is not None:
        data_lim = vlim
    elif not absolute:
        data_lim = [
                np.nanpercentile(data_fin, 100.0 - pct),
                np.nanpercentile(data_fin, pct)
                ]
    else:
        data_abs = np.absolute(data)
        data_lim_max = np.nanpercentile(data_abs, pct)
        data_lim = [-data_lim_max, data_lim_max]

    return data_lim


def colorbar(
        mappable, cax, clim, label=None, label_fontsize='large',
        numticks=5, tick_fontsize='large', tick_direction='in', tick_pad=4.0):
    """
    Plot colorbar on a given axis.

    """
    mappable.set_clim(*clim)
    cb = plt.colorbar(mappable, cax=cax)

    if isinstance(label, str):
        cb.set_label(label=label, fontsize=label_fontsize)

    cb.locator = mpl.ticker.LinearLocator(numticks=numticks)
    cb.ax.tick_params(
            labelsize=tick_fontsize,
            direction=tick_direction,
            pad=tick_pad)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Fitting kinematic moments.

@author: mathewvaridel
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.special import erf

from .const import PhysicalConstants

# constants
C = PhysicalConstants.C


class SpectralModel:

    def __init__(self, lines, lsf_fwhm, baseline_order=None, wave_ref=0.0):
        self.lines = lines
        self.nlines = len(lines)

        self.lsf_fwhm = lsf_fwhm
        self.lsf_sigma = lsf_fwhm/np.sqrt(8.0*np.log(2.0))

        self.baseline_order = baseline_order

        self.wave_ref = wave_ref

        self.nparam = self.nlines + 2
        if self.baseline_order is not None:
            self.nparam += 1 + self.baseline_order

    def calculate(self, wavelength, *param):
        if self.baseline_order is None:
            model = self._gas_model(wavelength, param)
        else:
            model = self._gas_model(
                    wavelength,
                    param[:self.nlines+2])
            model += self._baseline_model(
                    wavelength,
                    param[-1-self.baseline_order:],
                    )

        return model

    def fit_spaxel(self, wavelength, data, var=None, bounds=None):
        if bounds is None:
            # line flux bounds
            bounds = [
                    [np.log(1e-9)]*self.nlines,
                    [np.inf]*self.nlines,
                    ]

            # v, vdisp bounds
            bounds[0] += [-np.inf, np.log(1e-9)]
            bounds[1] += [np.inf, np.inf]

            if self.baseline_order is not None:
                bounds[0] += [-np.inf]*(self.baseline_order + 1)
                bounds[1] += [np.inf]*(self.baseline_order + 1)

        if (var is None) & np.any(data != 0.0):
            data_valid = np.isfinite(data)
            data_tmp = data[data_valid]
            w_tmp = wavelength[data_valid]
            sigma_tmp = None
        elif (var is None) & np.all(data == 0.0):
            popt = np.zeros(self.nparam)*np.nan
            pcov = np.zeros(self.nparam)*np.nan
            return popt, pcov
        elif np.any(var > 0.0):
            data_valid = ((var > 0.0) & np.isfinite(var) & np.isfinite(data))
            data_tmp = data[data_valid]
            w_tmp = wavelength[data_valid]
            sigma_tmp = np.sqrt(var[data_valid])
        else:
            popt = np.zeros(self.nparam)*np.nan
            pcov = np.zeros(self.nparam)*np.nan
            return popt, pcov

        if len(w_tmp) <= 1:
            popt = np.zeros(self.nparam)*np.nan
            pcov = np.zeros(self.nparam)*np.nan
            return popt, pcov

        try:
            guess = self._guess(w_tmp, data_tmp)
        except (ZeroDivisionError, IndexError):
            print(w_tmp, data_tmp)

        # enforce guess within bounds
        for i in range(len(guess)):
            if guess[i] < bounds[0][i]:
                guess[i] = bounds[0][i]
            elif guess[i] > bounds[1][i]:
                guess[i] = bounds[1][i]
            elif ~np.isfinite(guess[i]):
                guess[i] = 0.5*(bounds[1][i] - bounds[0][i])

        try:
            # print('GUESS', guess, bounds)
            # print(w_tmp, data_tmp, sigma_tmp)
            popt, pcov = curve_fit(
                    self.calculate,
                    w_tmp,
                    data_tmp,
                    sigma=sigma_tmp,
                    bounds=bounds,
                    p0=guess,
                    )

            pcov = pcov.diagonal().copy()

            # convert back to linear space
            popt[:self.nlines] = np.exp(popt[:self.nlines])
            popt[self.nlines+1] = np.exp(popt[self.nlines+1])
            pcov[:self.nlines] = np.exp(pcov[:self.nlines])
            pcov[self.nlines+1] = np.exp(pcov[self.nlines+1])
        except RuntimeError:
            # Occurs when curve_fit fails to converge
            popt = np.zeros(guess.size)*np.nan
            pcov = np.zeros(guess.size)*np.nan

        return popt, pcov

    def fit_cube(self, wavelength, data, var=None, wave_axis=2, **fit_kwargs):
        if wave_axis == 2:
            fit = np.zeros((self.nparam, *data.shape[:2]))*np.nan
            fit_err = np.zeros((self.nparam, *data.shape[:2]))*np.nan
            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    if var is not None:
                        fit[:, i, j], fit_err[:, i, j] = self.fit_spaxel(
                            wavelength, data[i, j, :], var[i, j, :],
                            **fit_kwargs)
                    else:
                        fit[:, i, j], fit_err[:, i, j] = self.fit_spaxel(
                            wavelength, data[i, j, :], **fit_kwargs)
        elif wave_axis == 0:
            fit = np.zeros((self.nparam, *data.shape[1:]))*np.nan
            fit_err = np.zeros((self.nparam, *data.shape[1:]))*np.nan
            for i in range(data.shape[1]):
                for j in range(data.shape[2]):
                    if var is not None:
                        fit[:, i, j], fit_err[:, i, j] = self.fit_spaxel(
                                wavelength, data[:, i, j], var[:, i, j],
                                **fit_kwargs)
                    else:
                        fit[:, i, j], fit_err[:, i, j] = self.fit_spaxel(
                                wavelength, data[:, i, j], **fit_kwargs)
        else:
            raise ValueError('Wave axis needs to be 0 or 2.')

        return fit, fit_err

    def _guess(self, wavelength, data, lambda_win=10.0):
        """Guess parameters using method of moments.

        data : array-like
        lambda_win : floatp
            Window to estimate parameters
        """
        dwave = wavelength[1] - wavelength[0]
        wave_left = wavelength - 0.5*dwave
        wave_right = wavelength + 0.5*dwave

        if self.baseline_order is None:
            guess = np.zeros(self.nlines + 2)
        else:
            guess = np.zeros(self.nlines + 2 + self.baseline_order + 1)

        tmp_flux = np.ones(self.nlines)*1e-9
        tmp_v = np.zeros(self.nlines)
        tmp_vdisp = np.ones(self.nlines)*1e-9*C/self.lines[0]

        guess[:self.nlines] = tmp_flux.copy()
        guess[self.nlines] = np.mean(tmp_v)
        guess[self.nlines+1] = np.mean(tmp_vdisp)
        for i, line in enumerate(self.lines):
            win = (
                (wave_right >= line[0] - lambda_win)
                & (wave_left <= line[0] + lambda_win)
                )
            if win.sum() <= 0:
                # if no valid pixels around emission line, then use initial
                # guesses
                continue

            win_data = data[win]
            win_wave = wavelength[win]

            tmp_flux[i] = max(1e-9, win_data.sum())
            guess[i] = max(1e-9, win_data.sum())

            weights = win_data - np.min(win_data)
            if np.all(weights == 0.0):
                weights = np.ones(weights.shape)

            mean_wave = np.average(win_wave, weights=weights)
            tmp_v[i] = (mean_wave/line[0] - 1.0)*C

            # TODO : Think below is wrong + it's a biased estimator
            tmp_vdisp[i] = np.sqrt(np.average(
                    (win_wave - mean_wave)**2, weights=weights))
            tmp_vdisp[i] = max(
                    1e-9,
                    np.sqrt(tmp_vdisp[i]**2 - self.lsf_sigma**2)
                    )
            tmp_vdisp[i] *= C/line[0]
            tmp_vdisp[i] = 20.0  # just for testing

        guess[self.nlines] = np.average(tmp_v, weights=tmp_flux)
        guess[self.nlines+1] = np.average(tmp_vdisp, weights=tmp_flux)

        # convert flux and vdisp to log
        guess[:self.nlines] = np.log(guess[:self.nlines])
        guess[self.nlines+1] = np.log(guess[self.nlines+1])

        return guess

    def _gas_model(self, wave, gas_param):
        model = np.zeros(len(wave))

        rel_lambda = 1.0 + gas_param[self.nlines]/C
        rel_lambda_sigma = np.exp(gas_param[self.nlines+1])/C

        # add emission line contribution
        for i, line in enumerate(self.lines):
            # model first line
            line_wave = line[0]
            line_flux = gas_param[i]

            lam = rel_lambda*line_wave
            lam_sigma = rel_lambda_sigma*line_wave

            model += self._gas_line_model(
                    wave, line_flux, lam, lam_sigma)

            nclines = len(line)//2
            for i in range(nclines):
                line_wave = line[1+2*i]
                factor = line[2+2*i]
                lam = rel_lambda*line_wave
                lam_sigma = rel_lambda_sigma*line_wave
                model += self._gas_line_model(
                        wave, factor*line_flux, lam, lam_sigma)

        return model

    def _gas_line_model(self, wave, flux, lam, lam_sigma):
        dwave = wave[1] - wave[0]
        wave_left = wave - 0.5*dwave
        wave_right = wave + 0.5*dwave

        var = lam_sigma**2 + self.lsf_sigma**2

        cdf_left = 0.5*erf((wave_left - lam)/np.sqrt(2.0*var))
        cdf_right = 0.5*erf((wave_right - lam)/np.sqrt(2.0*var))
        return np.exp(flux)*(cdf_right - cdf_left)  # /dx

    def _baseline_model(self, wave, baseline_param):
        wave_shft = wave - self.wave_ref
        baseline = np.ones(len(wave))*baseline_param[0]
        for i in range(self.baseline_order):
            baseline += baseline_param[1+i]*wave_shft**(1+i)
        return baseline


if __name__ == '__main__':
    import os
    from blobby3d import Blobby3D

    root = r'/Users/mathewvaridel/Google Drive/Code/Uni/Blobby3D/Examples/106717'
    b3d = Blobby3D(
        samples_path=os.path.join(root, 'posterior_sample.txt'),
        data_path=os.path.join(root, 'data.txt'),
        var_path=os.path.join(root, 'var.txt'),
        metadata_path=os.path.join(root, 'metadata.txt'),
        nlines=2)

    sm = SpectralModel(
            lines=[[6562.81], [6583.1, 6548.1, 0.3333]],
            lsf_fwhm=1.61
            )

    wave = np.linspace(b3d.r_lim[0] + 0.5*b3d.dr,
                       b3d.r_lim[1] - 0.5*b3d.dr,
                       b3d.naxis[2])

    i, j = 15, 15

    guess = sm._guess(wave, b3d.data[i, j, :])
    guess_model = sm._gas_model(wave, guess)

    fit, fit_err = sm.fit_spaxel(wave, b3d.data[i, j, :])
    fit_model = sm.calculate(wave, *fit)

    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(wave, guess_model)
    plt.plot(wave, fit_model)
    plt.plot(wave, b3d.data[i, j, :])

    fit_cube, fit_cube_err = sm.fit_cube(wave, b3d.data)
